@@ -4,7 +4,7 @@ package JSON::Schema::Modern::Document::OpenAPI;
 # ABSTRACT: One OpenAPI v3.1 or v3.2 document
 # KEYWORDS: JSON Schema data validation request response OpenAPI
 
-our $VERSION = '0.101';
+our $VERSION = '0.102';
 
 use 5.020;
 use utf8;
@@ -206,7 +206,7 @@ sub traverse ($self, $evaluator, $config_override = {}) {
 
   # evaluate the document against its metaschema to find any errors, to identify all schema
   # resources within to add to the global resource index, and to extract all operationIds
-  my (@json_schema_paths, @operation_paths, %bad_path_item_refs, @additional_operations_paths, @servers_paths);
+  my (@json_schema_paths, @operation_paths, %bad_path_item_refs, @servers_paths);
   my $result = $evaluator->evaluate(
     $schema, $self->metaschema_uri,
     {
@@ -242,9 +242,6 @@ sub traverse ($self, $evaluator, $config_override = {}) {
             $bad_path_item_refs{$state->{data_path}} = join(', ', sort keys %path_item) if keys %path_item;
           }
 
-          push @additional_operations_paths, $state->{data_path}.'/additionalOperations'
-            if $entity and $entity eq 'path-item' and exists $data->{additionalOperations};
-
           # will contain duplicates; filter out later
           push @servers_paths, ($state->{data_path} =~ s{/[0-9]+$}{}r)
             if $schema->{'$ref'} eq '#/$defs/server';
@@ -303,10 +300,6 @@ sub traverse ($self, $evaluator, $config_override = {}) {
       'invalid keywords used adjacent to $ref in a path-item: %s', $bad_path_item_refs{$path_item});
   }
 
-  foreach my $aop (sort @additional_operations_paths) {
-    ()= E({ %$state, schema_path => $aop }, 'not-yet-supported use of additionalOperations');
-  }
-
   my %seen_servers;
   foreach my $servers_location (reverse @servers_paths) {
     next if $seen_servers{$servers_location}++;
@@ -512,7 +505,7 @@ JSON::Schema::Modern::Document::OpenAPI - One OpenAPI v3.1 or v3.2 document
 
 =head1 VERSION
 
-version 0.101
+version 0.102
 
 =head1 SYNOPSIS
 
